@@ -18,6 +18,8 @@ package Dpkg::Source::Package::V2;
 use strict;
 use warnings;
 
+our $VERSION = "0.01";
+
 use base 'Dpkg::Source::Package';
 
 use Dpkg;
@@ -51,8 +53,14 @@ sub init_options {
         unless exists $self->{'options'}{'preparation'};
     $self->{'options'}{'skip_patches'} = 0
         unless exists $self->{'options'}{'skip_patches'};
+    $self->{'options'}{'unapply_patches'} = 0
+        unless exists $self->{'options'}{'unapply_patches'};
     $self->{'options'}{'skip_debianization'} = 0
         unless exists $self->{'options'}{'skip_debianization'};
+    $self->{'options'}{'create_empty_orig'} = 0
+        unless exists $self->{'options'}{'create_empty_orig'};
+    $self->{'options'}{'abort_on_upstream_changes'} = 0
+        unless exists $self->{'options'}{'abort_on_upstream_changes'};
 }
 
 sub parse_cmdline_option {
@@ -72,9 +80,18 @@ sub parse_cmdline_option {
     } elsif ($opt =~ /^--skip-patches$/) {
         $self->{'options'}{'skip_patches'} = 1;
         return 1;
+    } elsif ($opt =~ /^--unapply-patches$/) {
+        $self->{'options'}{'unapply_patches'} = 1;
+        return 1;
     } elsif ($opt =~ /^--skip-debianization$/) {
         $self->{'options'}{'skip_debianization'} = 1;
         return 1;
+    } elsif ($opt =~ /^--create-empty-orig$/) {
+        $self->{'options'}{'create_empty_orig'} = 1;
+        return 1;
+    } elsif ($opt =~ /^--abort-on-upstream-changes$/) {
+        $self->{'options'}{'abort_on_upstream_changes'} = 1;
+        return 1;
     }
     return 0;
 }
@@ -89,16 +106,17 @@ sub do_extract {
     my $basenamerev = $self->get_basename(1);
 
     my ($tarfile, $debianfile, %origtar, %seen);
+    my $re_ext = $compression_re_file_ext;
     foreach my $file ($self->get_files()) {
-        (my $uncompressed = $file) =~ s/\.$comp_regex$//;
+        (my $uncompressed = $file) =~ s/\.$re_ext$//;
         error(_g("duplicate files in %s source package: %s.*"), "v2.0",
               $uncompressed) if $seen{$uncompressed};
         $seen{$uncompressed} = 1;
-        if ($file =~ /^\Q$basename\E\.orig\.tar\.$comp_regex$/) {
+        if ($file =~ /^\Q$basename\E\.orig\.tar\.$re_ext$/) {
             $tarfile = $file;
-        } elsif ($file =~ /^\Q$basename\E\.orig-([\w-]+)\.tar\.$comp_regex$/) {
+        } elsif ($file =~ /^\Q$basename\E\.orig-([[:alnum:]-]+)\.tar\.$re_ext$/) {
             $origtar{$1} = $file;
-        } elsif ($file =~ /^\Q$basenamerev\E\.debian\.tar\.$comp_regex$/) {
+        } elsif ($file =~ /^\Q$basenamerev\E\.debian\.tar\.$re_ext$/) {
             $debianfile = $file;
         } else {
             error(_g("unrecognized file for a %s source package: %s"),
@@ -203,14 +221,41 @@ sub apply_patches {
     close(APPLIED);
 }
 
+sub unapply_patches {
+    my ($self, $dir, %opts) = @_;
+    my @patches = reverse($self->get_patches($dir, %opts));
+    return unless scalar(@patches);
+    my $timestamp = time();
+    my $applied = File::Spec->catfile($dir, "debian", "patches", ".dpkg-source-applied");
+    foreach my $patch (@patches) {
+        my $path = File::Spec->catfile($dir, "debian", "patches", $patch);
+        info(_g("unapplying %s"), $patch) unless $opts{"quiet"};
+        my $patch_obj = Dpkg::Source::Patch->new(filename => $path);
+        $patch_obj->apply($dir, force_timestamp => 1,
+                          timestamp => $timestamp,
+                          add_options => [ '-E', '-R' ]);
+    }
+    unlink($applied);
+}
+
 sub can_build {
     my ($self, $dir) = @_;
-    foreach ($self->find_original_tarballs()) {
-        return 1 if /\.orig\.tar\.$comp_regex$/;
-    }
+    return 1 if $self->find_original_tarballs(include_supplementary => 0);
+    return 1 if $self->{'options'}{'create_empty_orig'} and
+                $self->find_original_tarballs(include_main => 0);
     return (0, _g("no orig.tar file found"));
 }
 
+sub before_build {
+    my ($self, $dir) = @_;
+    $self->check_patches_applied($dir) if $self->{'options'}{'preparation'};
+}
+
+sub after_build {
+    my ($self, $dir) = @_;
+    $self->unapply_patches($dir) if $self->{'options'}{'unapply_patches'};
+}
+
 sub prepare_build {
     my ($self, $dir) = @_;
     $self->{'diff_options'} = {
@@ -221,6 +266,16 @@ sub prepare_build {
     };
     push @{$self->{'options'}{'tar_ignore'}}, "debian/patches/.dpkg-source-applied";
     $self->check_patches_applied($dir) if $self->{'options'}{'preparation'};
+    if ($self->{'options'}{'create_empty_orig'} and
+        not $self->find_original_tarballs(include_supplementary => 0))
+    {
+        # No main orig.tar, create a dummy one
+        my $filename = $self->get_basename() . ".orig.tar." .
+                       $self->{'options'}{'comp_ext'};
+        my $tar = Dpkg::Source::Archive->new(filename => $filename);
+        $tar->create();
+        $tar->finish();
+    }
 }
 
 sub check_patches_applied {
@@ -255,7 +310,7 @@ sub do_build {
     my ($tarfile, %origtar);
     my @origtarballs;
     foreach (sort $self->find_original_tarballs()) {
-        if (/\.orig\.tar\.$comp_regex$/) {
+        if (/\.orig\.tar\.$compression_re_file_ext$/) {
             if (defined($tarfile)) {
                 error(_g("several orig.tar files found (%s and %s) but only " .
                          "one is allowed"), $tarfile, $_);
@@ -263,7 +318,7 @@ sub do_build {
             $tarfile = $_;
             push @origtarballs, $_;
             $self->add_file($_);
-        } elsif (/\.orig-([\w-]+)\.tar\.$comp_regex$/) {
+        } elsif (/\.orig-([[:alnum:]-]+)\.tar\.$compression_re_file_ext$/) {
             $origtar{$1} = $_;
             push @origtarballs, $_;
             $self->add_file($_);
@@ -291,6 +346,7 @@ sub do_build {
     }
 
     # Copy over the debian directory
+    erasedir("$tmp/debian");
     system("cp", "-a", "--", "$dir/debian", "$tmp/");
     subprocerr(_g("copy of the debian directory")) if $?;
 
@@ -401,6 +457,9 @@ sub do_build {
                 syserr(_g("unable to change permission of `%s'"), $autopatch);
     }
     $self->register_autopatch($dir);
+    if (-e $autopatch and $self->{'options'}{'abort_on_upstream_changes'}) {
+        error(_g("aborting due to --abort-on-upstream-changes"));
+    }
     rmdir(File::Spec->catdir($dir, "debian", "patches")); # No check on purpose
     pop @Dpkg::Exit::handlers;
 
